import databases

from sqlalchemy import create_engine


class DatabaseManager:
    url = None
    database: databases.Database = None
    engine = None
    configured = False

    @classmethod
    async def install(cls, config):
        if cls.configured:
            return

        cls.url = (
            f"{config['type']}://"
            f"{config['username']}:{config['password']}@"
            f"{config['host']}:{config['port']}/"
            f"{config['database']}"
        )

        cls.engine = create_engine(cls.url)
        cls.database = databases.Database(cls.url)

        await cls.database.connect()

        cls.configured = True

    @classmethod
    async def disconnect(cls):
        await cls.database.disconnect()



from aioli import Package

from .service import DatabaseService
from .config import ConfigSchema


export = Package(
    controllers=[],
    services=[DatabaseService],
    config=ConfigSchema,
    auto_meta=True
)



from orm import models, fields


class ModelMeta(models.ModelMetaclass):
    def __new__(cls, name, bases, attrs):
        new_cls = super(models.ModelMetaclass, cls).__new__(
            cls, name, bases, attrs
        )

        mod_name = new_cls.__module__.split(".")[0]
        new_cls.__tablename__ = f"{mod_name}__{new_cls.__tablename__}"
        new_cls.__related__ = {}

        new_cls.columns = {}
        pkname = None

        for name, field in new_cls.fields.items():
            if field.primary_key:
                pkname = name

            new_cls.columns[name] = field.get_column(name)

        new_cls.__pkname__ = pkname

        return new_cls


class Model(models.Model, metaclass=ModelMeta):
    __metadata__ = None
    __tablename__ = None
    __table__ = None
    __registered__ = False



from enum import Enum

import orm

from sqlalchemy import select, func, desc, asc, text, sql

from aioli.exceptions import AioliException, NoMatchFound

from .manager import DatabaseManager


class FilterOperator(Enum):
    EXACT = "__eq__"
    IEXACT = "ilike"
    CONTAINS = "like"
    ICONTAINS = "ilike"
    IN = "in_"
    GT = "__gt__"
    GTE = "__ge__"
    LT = "__lt__"
    LTE = "__le__"


class ModelProxy:
    manager = DatabaseManager

    def __init__(self, model):
        self.model = model

    @property
    def related(self):
        return list(self.model.__related__.keys())

    def _model_has_attrs(self, *attrs):
        for attr in attrs:
            if attr not in self.model.fields:
                raise AioliException(400, f"Unknown field {attr}")

        return True

    def _parse_sortstr(self, value: str):
        if not value:
            return None

        for colname in value.split(","):
            sort_asc = True
            if colname.startswith("-"):
                colname = colname[1:]
                sort_asc = False

            if self._model_has_attrs(colname):
                # @TODO - support ordering by related fields
                tbl_colname = text(f"{self.model.__tablename__}.{colname}")
                yield asc(tbl_colname) if sort_asc else desc(tbl_colname)

    def _parse_query(self, **kwargs):
        # @TODO - implement _model_has_attrs for local and referenced values
        # @TODO - split up method, reuse in _parse_sortstr

        clauses = []

        for key, value in kwargs.items():
            if "__" in key:
                parts = key.split("__")

                op = parts[-1].upper()
                field_name = parts[-2]
                relations = parts[:-2]

                if len(relations) > 1:
                    raise AioliException(message="Unsupported query depth", status=400)
                elif len(relations) == 1:
                    related_name = relations[0]
                    column = self.model.__related__[related_name].columns[field_name]
                else:
                    column = self.model.__table__.columns[field_name]
            else:
                op = FilterOperator.EXACT.name
                column = self.model.__table__.columns[key]

            try:
                op_attr = FilterOperator[op].value
            except KeyError:
                raise AioliException(
                    message=f"Invalid operator: {op}, available: {[e.name for e in FilterOperator]}",
                    status=400,
                )

            if isinstance(value, orm.models.Model):
                value = value.pk
            elif op in [FilterOperator.CONTAINS.name, FilterOperator.ICONTAINS.name]:
                value = "%" + value + "%"

            clause = getattr(column, op_attr)(value)
            clauses.append(clause)

        if clauses:
            if len(clauses) == 1:
                return clauses[0]
            else:
                return sql.and_(*clauses)

        return None

    async def _get(self, **params):
        join = params.pop("join_related", True)
        limit = params.pop("limit", None)
        sort = params.pop("sort", None)
        offset = params.pop("offset", None)
        query = params.pop("query", None)

        related = self.related if join else []

        qs = orm.models.QuerySet(self.model, select_related=related)
        stmt = qs.build_select_expression().limit(limit).offset(offset)

        if query:
            if isinstance(query, str):
                clauses = [clause.split("=") for clause in query.split(",")]
                stmt = stmt.where(self._parse_query(**dict(clauses)))
            elif isinstance(query, dict):
                stmt = stmt.where(self._parse_query(**query))

        if sort:
            sort_fields = self._parse_sortstr(sort)
            stmt = stmt.order_by(*sort_fields)

        return [
            self.model.from_row(row, select_related=related)
            for row in await self.manager.database.fetch_all(stmt)
        ]

    async def get_many(self, **kwargs):
        return await self._get(**kwargs)

    async def get_one(self, **kwargs):
        pk_value = kwargs.pop("pk", None)
        if pk_value:
            kwargs[self.model.__pkname__] = pk_value

        result = await self._get(limit=2, query=kwargs)

        if len(result) == 0:
            raise NoMatchFound
        elif len(result) > 1:
            raise AioliException(message="Multiple matches for get_one", status=500)

        return result[0]

    async def create(self, **item: dict):
        return await orm.models.QuerySet(self.model).create(**item)

    async def count(self, **kwargs):
        clauses = self._parse_query(**kwargs)
        query = select([func.count()]).select_from(self.model.__table__).where(clauses)
        return await self.manager.database.fetch_val(query)

    async def update(self, record_id, payload):
        record = await self.get_one(pk=record_id)
        await record.update(**payload)
        return record

    async def delete(self, record_id):
        record = await self.get_one(pk=record_id)
        await record.delete()



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['aioli_rdbms']

package_data = \
{'': ['*']}

install_requires = \
['aioli>=0.1,<0.2',
 'aiomysql>=0.0.20,<0.0.21',
 'asyncpg>=0.18.3,<0.19.0',
 'databases>=0.2.1,<0.3.0',
 'mysqlclient>=1.4.2,<2.0.0',
 'orm>=0.1,<0.2',
 'psycopg2-binary>=2.8.1,<3.0.0']

setup_kwargs = {
    'name': 'aioli-rdbms',
    'version': '0.1.6',
    'description': 'ORM and CRUD Service for Aioli with support for MySQL and PostgreSQL',
    'long_description': '',
    'author': 'Robert Wikman',
    'author_email': 'rbw@vault13.org',
    'url': 'https://github.com/aioli-framework/aioli-rdbms',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from aioli.config import PackageConfigSchema, fields, validate


class ConfigSchema(PackageConfigSchema):
    type = fields.String(
        validate=validate.OneOf(["mysql", "postgres"]),
        required=True
    )
    username = fields.String(required=True)
    password = fields.String(required=True)
    host = fields.String(missing="127.0.0.1")
    port = fields.Integer(missing=3306)
    database = fields.String(missing="aioli")



from sqlalchemy import MetaData, Table

from aioli.service import BaseService

from .model import fields
from .proxy import ModelProxy
from .manager import DatabaseManager


class DatabaseService(BaseService):
    _models_registered = []

    async def on_startup(self):
        mgr = DatabaseManager

        if not mgr.configured:
            await mgr.install(self.config)

        for model in self._models_registered:
            model.__database__ = mgr.database

        if self.pkg.stash["db_metadata"]:
            self.pkg.stash["db_metadata"].create_all(mgr.engine)

    async def on_shutdown(self):
        self.log.debug("Disconnecting from database")
        await DatabaseManager.disconnect()

    def use_model(self, model):
        """Factory method for creating a ModelProxy

        :param model: Model class
        :return: ModelProxy instance
        """

        if model.__registered__ is False:
            model = self._register_model(model)

        return ModelProxy(model)

    def _register_model(self, model):
        assert model.__registered__ is False, f"Model {model} has already been registered"

        stash = self.pkg.stash

        if not stash["db_metadata"]:
            stash["db_metadata"] = MetaData()

        self.log.info(f"Registering Model: {model.__name__} [{model.__tablename__}]")

        for name, field in model.fields.items():
            if isinstance(field, fields.ForeignKey):
                related_mdl = field.to
                if related_mdl.__registered__ is False:
                    self._register_model(related_mdl)

                model.__related__[name] = related_mdl

        model.__metadata__ = stash["db_metadata"]
        model.__table__ = Table(model.__tablename__, stash["db_metadata"], *model.columns.values())
        model.__registered__ = True

        self._models_registered.append(model)

        return model
